/*#####################################################################
  # Copyright (C) 2000 Artur Trzewik <mail@xdobry.de>
  #
  # This program is free software; you can red^istribute it and/or
  # modify it under the terms of the GNU General Public License
  # as published by the Free Software Foundation; either version 2
  # of the License, or (at your option) any later version.
  #
  # This program is distributed in the hope that it will be useful,
  # but WITHOUT ANY WARRANTY; without even the implied warranty of
  # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  # GNU General Public License for more details.
  ######################################################################*/

#include <EXTERN.h>
#include <perl.h>
#include <XSUB.h>
#include <tcl.h>

typedef struct dbitotcl_state {
  PerlInterpreter *myperl;
  int conn;
  Tcl_Obj *null_obj;
} dbitotcl_state;
typedef struct connect_data {
  dbitotcl_state *state;
  SV *hash;
  int count;
  int result;
} connect_data;
typedef struct result_data {
  connect_data *connect;
  SV *hash;
} result_data;
/*
Transfer Perl SV to Tcl Object
*/
Tcl_Obj *
sv_to_tclobj(SV *val,Tcl_Interp  *interp) {
  STRLEN n_a;
  I32   len,i;
  Tcl_Obj *obj,*hpair;
  SV *elem;
  MAGIC *mg; 
  char *key;
  //printf ("sv_to_tclobj %d val=%s\n",SvTYPE(val),SvPV(val,n_a));
  switch SvTYPE(val) {
  case  SVt_IV: {
    return Tcl_NewIntObj(SvIV(val));
  }
  case  SVt_NV: {
    return Tcl_NewDoubleObj(SvNV(val));
  }
  case  SVt_PV: {
    return Tcl_NewStringObj(SvPV(val,n_a),-1);
  }
  case  SVt_PVAV: {
    obj=Tcl_NewListObj(0, NULL);
    for (i=0;i<=av_len((AV*)val);i++) {
      elem=*av_fetch((AV*)val,i,1);
      Tcl_ListObjAppendElement (interp, obj,sv_to_tclobj(elem,interp));
    }
    return obj;
  }
  case  SVt_PVHV: {
    obj=Tcl_NewListObj(0, NULL);
    hv_iterinit((HV*)val);
    while( (elem = hv_iternextsv((HV*)val, &key, &len)) ) {
      hpair=Tcl_NewListObj(0, NULL);
      Tcl_ListObjAppendElement (interp, hpair,Tcl_NewStringObj(key,-1));
      Tcl_ListObjAppendElement (interp, hpair,sv_to_tclobj(elem,interp));
      Tcl_ListObjAppendElement (interp, obj,hpair);
    }
    return obj;
  }
  case SVt_RV: {
    return sv_to_tclobj(SvRV(val),interp);
  }
  case SVt_PVMG: {
    /* I do not understand the magic variables, there are some by tied arrays or hashes */
/*      //mg_get(val); */
/*      if (mg=mg_find(val,'p')) { */
/*        elem=mg->mg_obj; */
/*        return sv_to_tclobj(elem,interp); */
/*      }  */
    mg_get(val);
    elem=SvRV(val);
    //mg_get(elem);
    return sv_to_tclobj(elem,interp);
  }
  }  
  return Tcl_NewStringObj(SvPV(val,n_a),-1);
}
/* 
read Attribute from object and
transfer to Tcl
use FETCH method
*/
static int 
read_attribute(SV *hash,Tcl_Interp  *interp,int objc,Tcl_Obj *CONST objv[]) {
  dSP ;
  int count,i,retvalue=TCL_OK;
  Tcl_Obj *ret;
  SV *res;
  I32 ax ;
  if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "attribute");
      return TCL_ERROR;
  }
  ENTER ;
  SAVETMPS ;
  PUSHMARK(SP) ;
  XPUSHs(hash);
  XPUSHs(sv_2mortal(newSVpv(Tcl_GetStringFromObj(objv[2],0),0)));
  PUTBACK ;
  count=perl_call_method("FETCH", G_ARRAY);
  SPAGAIN ;
  SP -= count ;
  ax = (SP - PL_stack_base) + 1 ;
  if (count>0) {
    if (count>1) {
      ret=Tcl_NewListObj(0, NULL);
      for (i=0;i<count;i++) {
	res= ST(i);
	Tcl_ListObjAppendElement (interp, ret,sv_to_tclobj(res,interp));
      }
    } else {
      res= ST(0);
      if (SvTRUE(res)) {
	ret=sv_to_tclobj(res,interp);
      } else {
	ret=Tcl_NewStringObj("can not fetch the attribute",-1);	
	retvalue=TCL_ERROR;
      }
    }
    Tcl_SetObjResult(interp,ret);
  }     
  PUTBACK ;
  FREETMPS ;
  LEAVE ;
  return retvalue;
/*   SV **val; */
/*   STRLEN n_a; */
/*   char *attr; */
/*   int len; */
/*   Tcl_Obj *obj; */
/*   if (objc != 3) { */
/*       Tcl_WrongNumArgs(interp, 2, objv, "attribute"); */
/*       return TCL_ERROR; */
/*   } */
/*   if (SvROK(hash) && SvTYPE(SvRV(hash))==SVt_PVHV) { */
/*     attr=Tcl_GetStringFromObj(objv[2],&len); */
/*     if (hv_exists((HV*)SvRV(hash),attr,len)) { */
/*       val=hv_fetch((HV*)SvRV(hash),attr,len,0); */
/*       if (val) { */
/* 	obj=sv_to_tclobj(*val,interp); */
/* 	if (obj) { */
/* 	  Tcl_SetObjResult(interp,obj); */
/* 	} else { */
/* 	  Tcl_SetObjResult(interp,Tcl_NewStringObj(SvPV(*val,n_a),-1)); */
/* 	} */
/*       } */
/*     } else { */
/*       Tcl_SetObjResult(interp, Tcl_NewStringObj("attribute not found", -1));   */
/*       return TCL_ERROR; */
/*     } */
/*   } else { */
/*     panic("internal Error dbitotcl readattribute"); */
/*   } */
/*   return TCL_OK; */
}
static void
set_db_error_info (Tcl_Interp *interp,SV *hash) {
  dSP ;
  ENTER ;
  SAVETMPS ;
  
  PUSHMARK(SP) ;
  XPUSHs(hash);
  PUTBACK ;

  perl_call_method("errstr", G_SCALAR);

  SPAGAIN ;

  Tcl_SetObjResult(interp,Tcl_NewStringObj(POPp,-1));
  
  PUTBACK ;
  FREETMPS ;
  LEAVE ;
}
static void
dbiset_db_error (Tcl_Interp *interp) {
  SV *dbih;
  dbih=newSVpv("DBI",0);
  set_db_error_info(interp,dbih);
  SvREFCNT_dec(dbih); 
}
/* 
invoke method (withount attributes) on object
if returned value was UNDEF set return to TCL_ERROR and
ObjResult to hash->errstr
*/
static int 
invoke_method_err(Tcl_Interp  * interp,SV *hash,char *mstring) {
  dSP ;
  int ret=TCL_OK;
  SV *res;
  ENTER ;
  SAVETMPS ;
  PUSHMARK(SP) ;
  XPUSHs(hash);
  PUTBACK ;
  perl_call_method(mstring, G_SCALAR);
  SPAGAIN ;
  res=POPs;
  if (!SvTRUE(res)) {
    ret=TCL_ERROR;
  }
  PUTBACK ;
  FREETMPS ;
  LEAVE ;
  if (ret==TCL_ERROR) {
    set_db_error_info(interp,hash);
  }
  return ret;
}
/* 
invoke method (withount attributes) on object
handle returned Value As Tcl List 
*/
static int 
invoke_method_retlist(Tcl_Interp  *interp,SV *hash,char *mstring) {
  dSP ;
  int count,i;
  Tcl_Obj *ret;
  SV *res;
  I32 ax ;
  ENTER ;
  SAVETMPS ;
  PUSHMARK(SP) ;
  XPUSHs(hash);
  PUTBACK ;
  count=perl_call_method(mstring, G_ARRAY);
  SPAGAIN ;
  SP -= count ;
  ax = (SP - PL_stack_base) + 1 ;
  if (count>0) {
    ret=Tcl_NewListObj(0, NULL);
    for (i=0;i<count;i++) {
      res= ST(i);
      Tcl_ListObjAppendElement (interp, ret,
				Tcl_NewStringObj(SvPV(res,PL_na),-1));
    }
    Tcl_SetObjResult(interp,ret);
  }     
  PUTBACK ;
  FREETMPS ;
  LEAVE ;
  return TCL_OK;
}
/* 
invoke mehtod on diffrents objects (hash)
all Attributes are loaded on the Perl Stack
returned Value As Tcl List 
*/
static int 
invoke_func(SV *hash,Tcl_Interp  *interp,int objc,Tcl_Obj *CONST objv[]) {
  dSP ;
  int count,i;
  Tcl_Obj *ret;
  SV *res;
  I32 ax ;
  printf("argc %d method %s\n",objc,Tcl_GetStringFromObj(objv[2],0));
  if (objc < 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "method ?arg1? ?arg2? ....");
      return TCL_ERROR;
  }
  ENTER ;
  SAVETMPS ;
  PUSHMARK(SP) ;
  XPUSHs(hash);
  for (i=3;i<objc;i++) {
    XPUSHs(sv_2mortal(newSVpv(Tcl_GetStringFromObj(objv[i],0),0)));
  }
  PUTBACK ;
  count=perl_call_method(Tcl_GetStringFromObj(objv[2],0), G_ARRAY);
  SPAGAIN ;
  SP -= count ;
  ax = (SP - PL_stack_base) + 1 ;
  if (count>0) {
    if (count>1) {
      ret=Tcl_NewListObj(0, NULL);
      for (i=0;i<count;i++) {
	res= ST(i);
	Tcl_ListObjAppendElement (interp, ret,sv_to_tclobj(res,interp));
      }
    } else {
      res= ST(0);
      ret=sv_to_tclobj(res,interp);
    }
    Tcl_SetObjResult(interp,ret);
  }     
  PUTBACK ;
  FREETMPS ;
  LEAVE ;
  return TCL_OK;
}

static int
result_finish (ClientData clientData,Tcl_Interp  * interp,
		     int objc,Tcl_Obj *CONST objv[]) {
  result_data *result=(result_data *)clientData;
  int ret;
  if (objc != 2) {
      Tcl_WrongNumArgs(interp, 2, objv, "");
      return TCL_ERROR;
  }
  ret=invoke_method_err(interp,result->hash,"finish");
  Tcl_DeleteCommand(interp,Tcl_GetStringFromObj(objv[0],0));
  SvREFCNT_dec(result->hash);
  Tcl_Free((char *)result);
  return ret;
}
static int
result_fetchrow (ClientData clientData,Tcl_Interp  * interp,
		     int objc,Tcl_Obj *CONST objv[]) {
  dSP ;
  result_data *result=(result_data *)clientData;
  SV *res;
  Tcl_Obj *ret;
  int count,i;
  I32 ax ;
  if (objc != 2) {
      Tcl_WrongNumArgs(interp, 2, objv, "");
      return TCL_ERROR;
  }
  ENTER ;
  SAVETMPS ;
  
  PUSHMARK(SP) ;
  XPUSHs(result->hash);
  PUTBACK ;

  count = perl_call_method("fetchrow_array", G_ARRAY);

  SPAGAIN ;
  SP -= count ;
  ax = (SP - PL_stack_base) + 1 ;
   
  if (count>0) {
    ret=Tcl_NewListObj(0, NULL);
    for (i=0;i<count;i++) {
      res= ST(i);
      if (SvTRUE(res)) {
	/* parhaps better Tcl_ExternalToUtf or ByteArrays */
	Tcl_ListObjAppendElement (interp, ret,
				  Tcl_NewStringObj(SvPV(res,PL_na),-1));
      } else {
	Tcl_ListObjAppendElement (interp, ret,Tcl_DuplicateObj(result->connect->state->null_obj));
      }
    }
    Tcl_SetObjResult(interp,ret);
  }     

  PUTBACK ;
  FREETMPS ;
  LEAVE ;

  return TCL_OK;
}
static int
handle_Result (ClientData clientData,Tcl_Interp  * interp,
		     int objc,Tcl_Obj *CONST objv[]) {
  result_data *result=(result_data *)clientData;
  static char *commandStrings[] = {
    "execute","finish","fetchrow","invoke","attribute",NULL
  };
  int index;
  if (Tcl_GetIndexFromObj(interp, objv[1], commandStrings, "command", TCL_EXACT, &index)==TCL_ERROR) {
    return TCL_ERROR;
  }
  switch (index) {
  case 0: {
    if (objc != 2) {
      Tcl_WrongNumArgs(interp, 2, objv, "");
      return TCL_ERROR;
    }
    return invoke_method_err(interp,result->hash,"execute");
    break;
  }
  case 1: {
    return result_finish(clientData,interp,objc,objv);
    break;
  }
  case 2: {
    return result_fetchrow(clientData,interp,objc,objv);
    break;
  }
  case 3: {
    return invoke_func(result->hash,interp,objc,objv);
  }
  case 4: {
    return read_attribute(result->hash,interp,objc,objv);
  }
  }
}
static int
conn_disconnect (ClientData clientData,Tcl_Interp  * interp,
		     int objc,Tcl_Obj *CONST objv[]) {
  connect_data *conn=(connect_data *)clientData;
  int ret;
  if (objc != 2) {
      Tcl_WrongNumArgs(interp, 2, objv, "");
      return TCL_ERROR;
  }
  ret=invoke_method_err(interp,conn->hash,"disconnect");

  Tcl_DeleteCommand(interp,Tcl_GetStringFromObj(objv[0],0));
  SvREFCNT_dec(conn->hash);
  Tcl_Free((char *)conn);
  return ret;
}
static int 
conn_prepare (ClientData clientData,Tcl_Interp  * interp,
		  int objc,Tcl_Obj *CONST objv[]) {
  dSP;
  char buf[40];
  Tcl_DString query;
  int count;
  connect_data *conn=(connect_data *)clientData;
  result_data *result;
  if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "sqlstatment");
      return TCL_ERROR;
  }
  ENTER ;
  SAVETMPS ;
  
  PUSHMARK(SP) ;
  XPUSHs(conn->hash);
#if (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION>0) || TCL_MAJOR_VERSION>8
  Tcl_UtfToExternalDString(NULL, Tcl_GetStringFromObj(objv[2],0), -1, &query);
  XPUSHs(sv_2mortal(newSVpv(Tcl_DStringValue(&query),0)));
  Tcl_DStringFree(&query);  
#else
  XPUSHs(sv_2mortal(newSVpv(Tcl_GetStringFromObj(objv[2],0),0)));
#endif
  PUTBACK ;

  count=perl_call_method("prepare", G_SCALAR);
  
  SPAGAIN ;

  result=(result_data *)Tcl_Alloc(sizeof(result_data));
  result->hash=newSVsv(POPs);

  PUTBACK ;
  FREETMPS ;
  LEAVE ;

  if (SvTRUE(result->hash)) {
    result->connect=conn;
    sprintf(buf,"result%d_%d",conn->count,conn->result++);
    Tcl_CreateObjCommand (interp, buf, handle_Result, 
			  (ClientData)result,NULL);
    Tcl_SetObjResult(interp, Tcl_NewStringObj(buf, -1));
    return TCL_OK;
  } else {
    SvREFCNT_dec(result->hash);
    Tcl_Free((char *)result);
    set_db_error_info(interp,conn->hash);
    return TCL_ERROR;
  }
}
static int 
conn_table_info (ClientData clientData,Tcl_Interp  * interp,
		  int objc,Tcl_Obj *CONST objv[]) {
  dSP;
  char buf[40];
  int count;
  connect_data *conn=(connect_data *)clientData;
  result_data *result;
  if (objc != 2) {
      Tcl_WrongNumArgs(interp, 2, objv, "");
      return TCL_ERROR;
  }
  ENTER ;
  SAVETMPS ;
  
  PUSHMARK(SP) ;
  XPUSHs(conn->hash);
  PUTBACK ;

  count=perl_call_method("table_info", G_SCALAR);
  
  SPAGAIN ;

  result=(result_data *)Tcl_Alloc(sizeof(result_data));
  result->hash=newSVsv(POPs);

  PUTBACK ;
  FREETMPS ;
  LEAVE ;

  if (SvTRUE(result->hash)) {
    result->connect=conn;
    sprintf(buf,"aresult%d_%d",conn->count,conn->result++);
    Tcl_CreateObjCommand (interp, buf, handle_Result, 
			  (ClientData)result,NULL);
    Tcl_SetObjResult(interp, Tcl_NewStringObj(buf, -1));
    return TCL_OK;
  } else {
    SvREFCNT_dec(result->hash);
    Tcl_Free((char *)result);
    set_db_error_info(interp,conn->hash);
    return TCL_ERROR;
  }
}
static int 
conn_do (ClientData clientData,Tcl_Interp  * interp,
	     int objc,Tcl_Obj *CONST objv[]) {
  dSP ;
  Tcl_DString query;
  SV *res;
  int ret=TCL_OK;
  connect_data *conn=(connect_data *)clientData;
  if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "sqlstatment");
      return TCL_ERROR;
  }
  ENTER ;
  SAVETMPS ;
  
  PUSHMARK(SP) ;
  XPUSHs(conn->hash);
#if (TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION>0) || TCL_MAJOR_VERSION>8
  Tcl_UtfToExternalDString(NULL, Tcl_GetStringFromObj(objv[2],0), -1, &query);
  XPUSHs(sv_2mortal(newSVpv(Tcl_DStringValue(&query),0)));
  Tcl_DStringFree(&query);  
#else
  XPUSHs(sv_2mortal(newSVpv(Tcl_GetStringFromObj(objv[2],0),0)));
#endif
  PUTBACK ;

  perl_call_method("do", G_SCALAR);

  SPAGAIN ;

  res=POPs;
  if (SvTRUE(res)) {
    Tcl_SetObjResult(interp,Tcl_NewIntObj(POPi));
  } else {
    set_db_error_info(interp,conn->hash);
    ret=TCL_ERROR;
  }
  
  PUTBACK ;
  FREETMPS ;
  LEAVE ;
  return ret;
}
static int 
conn_quote (ClientData clientData,Tcl_Interp  * interp,
	     int objc,Tcl_Obj *CONST objv[]) {
  dSP ;
  connect_data *conn=(connect_data *)clientData;
  int ret=TCL_OK;
  SV *res;
  if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "string");
      return TCL_ERROR;
  }
  ENTER ;
  SAVETMPS ;
  PUSHMARK(SP) ;
  XPUSHs(conn->hash);
  XPUSHs(sv_2mortal(newSVpv(Tcl_GetStringFromObj(objv[2],0),0)));
  PUTBACK ;
  perl_call_method("quote", G_SCALAR);
  SPAGAIN ;
  Tcl_SetObjResult(interp,Tcl_NewStringObj(POPp,-1));  
  PUTBACK ;
  FREETMPS ;
  LEAVE ;
  return TCL_OK;
}

static int 
handle_Connection (ClientData clientData,Tcl_Interp  * interp,
		       int objc,Tcl_Obj *CONST objv[]) {
  connect_data *conn=(connect_data *)clientData;
  static char *commandStrings[] = {
    "disconnect","prepare","do","quote","invoke","attribute",
    "table_info",NULL
  };
  int index;
  if (Tcl_GetIndexFromObj(interp, objv[1], commandStrings, "command", TCL_EXACT, &index)==TCL_ERROR) {
    return TCL_ERROR;
  }
  switch (index) {
  case 0: {
    return conn_disconnect(clientData,interp,objc,objv);
    break;
  }
  case 1: {
    return conn_prepare(clientData,interp,objc,objv);
    break;
  }
  case 2: {
    return conn_do(clientData,interp,objc,objv);
    break;
  }
  case 3: {
    return conn_quote(clientData,interp,objc,objv);
    break;
  }
  case 4: {
    return invoke_func(conn->hash,interp,objc,objv);
  }
  case 5: {
    return read_attribute(conn->hash,interp,objc,objv);
  }
  case 6: {
    return conn_table_info(clientData,interp,objc,objv);
  }
  }
}
static int 
dbitotcl_Connect (clientData, interp, objc, objv)
     ClientData   clientData;
     Tcl_Interp  *interp;
     int          objc;
     Tcl_Obj *CONST objv[];
{
  dSP ;
  int count;
  connect_data *conn;
  char buf[30];
  dbitotcl_state *state=(dbitotcl_state *)clientData;
  if (objc != 3 && objc !=4) {
      Tcl_WrongNumArgs(interp, 1, objv, "DBIstring user ?password?");
      return TCL_ERROR;
  }
  
  ENTER ;
  SAVETMPS;

  PUSHMARK(SP) ;  
  XPUSHs(sv_2mortal(newSVpv("DBI",0)));
  XPUSHs(sv_2mortal(newSVpv(Tcl_GetStringFromObj(objv[1],0),0)));
  XPUSHs(sv_2mortal(newSVpv(Tcl_GetStringFromObj(objv[2],0),0)));
  if (objc==4) 
    XPUSHs(sv_2mortal(newSVpv(Tcl_GetStringFromObj(objv[3],0),0)));
  PUTBACK ;

  count = perl_call_method("connect", G_SCALAR);
  
  SPAGAIN ;

  conn=(connect_data *)Tcl_Alloc(sizeof(connect_data));
  conn->hash=newSVsv(POPs);
  conn->result=0;

  PUTBACK ;
  FREETMPS ;
  LEAVE ;

  if (SvTRUE(conn->hash)) {
    conn->count=state->conn++;
    conn->state=state;
    sprintf(buf,"dbiconn%d",conn->count);
    Tcl_CreateObjCommand (interp, buf, handle_Connection, 
			  (ClientData)conn,NULL);
    Tcl_SetObjResult(interp, Tcl_NewStringObj(buf, -1));
    return TCL_OK;  
  } else {
    SvREFCNT_dec(conn->hash);
    Tcl_Free((char *)conn);
    dbiset_db_error(interp);
    return TCL_ERROR;
  }
}
/* static int  */
/* dbitotcl_attribute (ClientData clientData,Tcl_Interp  * interp, */
/* 		       int objc,Tcl_Obj *CONST objv[]) { */
/*   int ret; */
/*   SV *dbih; */
/*   if (objc != 2) { */
/*       Tcl_WrongNumArgs(interp, 1, objv, "attribute"); */
/*       return TCL_ERROR; */
/*   } */
/*   dbih=newSVpv("DBI",0); */
/*   ret=read_attribute(dbih,interp,objc+1,--objv); */
/*   SvREFCNT_dec(dbih);  */
/*   return ret; */
/* } */
static int 
dbitotcl_invoke (ClientData clientData,Tcl_Interp  * interp,
		       int objc,Tcl_Obj *CONST objv[]) {
  int ret;
  SV *dbih;
  if (objc < 2) {
      Tcl_WrongNumArgs(interp, 1, objv, "method ?arg1? ?arg2? ....");
      return TCL_ERROR;
  }
  dbih=newSVpv("DBI",0);
  ret=invoke_func(dbih,interp,objc+1,--objv);
  SvREFCNT_dec(dbih); 
  return ret;
}
static int 
dbitotcl_setnull (ClientData clientData,Tcl_Interp  * interp,
		       int objc,Tcl_Obj *CONST objv[]) {
  dbitotcl_state *state=(dbitotcl_state *)clientData;
  if (objc != 2) {
      Tcl_WrongNumArgs(interp, 1, objv, "nullvalue");
      return TCL_ERROR;
  }
  Tcl_DecrRefCount(state->null_obj);
  state->null_obj=Tcl_DuplicateObj(objv[1]);
  return TCL_OK;
}
static void dbitotcl_destroy (ClientData clientData) {
  dbitotcl_state *state=(dbitotcl_state *)clientData;
  if (state->myperl) {
    perl_destruct(state->myperl);
    perl_free(state->myperl);
    state->myperl=NULL;
  }
}

extern void xs_init _((void));
int Dbitotcl_Init (Tcl_Interp *interp) {
  char *enbedding[] = { "", "-e", "0" };
  dbitotcl_state *state=(dbitotcl_state *)Tcl_Alloc(sizeof(dbitotcl_state));
  SV *command;

  state->myperl=perl_alloc();
  state->conn=0;
  /* Init the Object Used as NULL Value, returned from PERL as undef */
  state->null_obj=Tcl_NewStringObj("NULL",-1);
  perl_construct(state->myperl);
  perl_parse(state->myperl, xs_init, 3, enbedding, NULL);
  perl_run(state->myperl);

  command=newSVpv("use DBI;", 0);
  perl_eval_sv(command, G_VOID|G_DISCARD); 

  Tcl_CreateObjCommand (interp, "dbi::connect", dbitotcl_Connect, 
		       (ClientData)state,dbitotcl_destroy);
  Tcl_CreateObjCommand (interp, "dbi::invoke",
			dbitotcl_invoke,
		       (ClientData)state,dbitotcl_destroy);
/*   Tcl_CreateObjCommand (interp, "dbi::attribute", */
/* 			dbitotcl_attribute, */
/* 		       (ClientData)state,dbitotcl_destroy); */
  Tcl_CreateObjCommand (interp, "dbi::setnull",
			dbitotcl_setnull,
		       (ClientData)state,dbitotcl_destroy);


  if (Tcl_PkgProvide(interp, "dbitotcl", "0.1") != TCL_OK)
    return TCL_ERROR;
  return TCL_OK;
}
